from pathlib import Path
from typing import Dict, Union
from types import SimpleNamespace

import cv2
import torch
from torch import nn

from fdfat.nn import model
from fdfat.utils.logger import LOGGER
from fdfat.cfg import get_cfg, get_cfg_data, yaml_save, cfg2dict, yaml_print
from fdfat.utils.utils import LMK_PARTS, LMK_PART_NAMES, increment_path, generate_graph, read_file_list, render_lmk
from fdfat.utils.model_utils import init_seeds, select_device, model_info

class BaseEngine:

    def __init__(self, cfg_: Union[str, Path, Dict, SimpleNamespace]):

        if isinstance(cfg_, (str, Path)):
            self.cfgs = get_cfg(cfg_)  # load dict
        elif isinstance(cfg_, (Dict)):
            self.cfgs = SimpleNamespace(**vars(cfg_))
        else:
            self.cfgs = cfg_

        init_seeds(self.cfgs.seed)
        # cv2.setNumThreads(0)  # prevent OpenCV from multithreading

        self.save_dir = Path(increment_path(Path(self.cfgs.project) / self.cfgs.name, exist_ok=self.cfgs.task != "train", mkdir=(self.cfgs.task == "train" and not self.cfgs.resume)))
        LOGGER.info(f"Project path: {self.save_dir}")
        self.save_wdir = self.save_dir / 'weights'  # weights dir
        self.save_wdir.mkdir(parents=True, exist_ok=True)  # make directory
        self.save_last, self.save_best = self.save_wdir / 'last.pt', self.save_wdir / 'best.pt'  # checkpoint paths
        self.save_log_csv = self.save_dir / 'log.csv'
        self.save_log_png = self.save_dir / 'log.png'
        self.save_config = self.save_dir / 'config.yaml'
        if self.cfgs.task == "train":
            yaml_save(self.save_config, cfg2dict(self.cfgs))

        yaml_print(self.cfgs)

        self.cfgs.save_dir = self.save_dir
        if self.cfgs.data is not None:
            self.cfgs.data = get_cfg_data(self.cfgs.data)
            if self.cfgs.cache:
                self.cfgs.data.train_cache = self.cfgs.data.train.replace(".txt", ".cache")
                self.cfgs.data.val_cache = self.cfgs.data.val.replace(".txt", ".cache")
                self.cfgs.data.test_cache = self.cfgs.data.test.replace(".txt", ".cache")
            else:
                self.cfgs.data.train_cache = None
                self.cfgs.data.val_cache = None
                self.cfgs.data.test_cache = None

        self.cfgs.device = select_device(self.cfgs.device)

        LOGGER.info(f"Using {self.cfgs.device} device")

        self.start_epoch = 0
        self.best_epoch_loss = 999
        self.best_epoch_no = 0

        if self.cfgs.lmk_num == 70:
            self.cfgs.lmk_parts = LMK_PARTS
            self.cfgs.lmk_part_names = LMK_PART_NAMES
        else: # 68 points, no purpil
            self.cfgs.lmk_parts = LMK_PARTS[:-1]
            self.cfgs.lmk_part_names = LMK_PART_NAMES[:-1]

    def load_database(self):
        raise NotImplementedError("Not implemented")
    
    def prepare(self):
        raise NotImplementedError("Not implemented")

    def load_model(self, verbose=True):
        LOGGER.info(f"Load model: {self.cfgs.model}")
        self.net = getattr(model, self.cfgs.model)(imgz=self.cfgs.imgsz, muliplier=self.cfgs.muliplier, pose_rotation=self.cfgs.aux_pose).to(self.cfgs.device)
        if verbose:
            # need to call this before getting model info to prevent abnormal batchnorm
            self.net.eval()
            _ = model_info(self.net, detailed=True, imgsz=self.cfgs.imgsz, device=self.cfgs.device)

    def load_checkpoint(self, checkpoint, map_location='cpu', epoch_info=True):
        if isinstance(checkpoint, str) or isinstance(checkpoint, Path):
            checkpoint = torch.load(checkpoint, map_location=map_location)
        self.checkpoint = checkpoint
        self.net.load(checkpoint)
        if epoch_info:
            self.start_epoch = checkpoint['epoch']
            self.best_epoch_loss = checkpoint['best_fit']
            self.best_epoch_no = checkpoint['best_epoch']
        LOGGER.info(f"Loaded checkpoint epoch {checkpoint['epoch']}")


import torch
from torch import nn

from fdfat.utils.model_utils import intersect_dicts
from fdfat.utils.logger import LOGGER

from .conv import *
from . import module
from . import module2
from . import module3
from . import module4
from . import module5

class BaseModel(nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x
    
    def is_fused(self):
        return False
    
    def load(self, weights, verbose=True):
        """Load the weights into the model.

        Args:
            weights (dict) or (torch.nn.Module): The pre-trained weights to be loaded.
            verbose (bool, optional): Whether to log the transfer progress. Defaults to True.
        """
        model = weights['model'] if isinstance(weights, dict) else weights  # torchvision models are not dicts
        csd = model.float().state_dict()  # checkpoint state_dict as FP32
        csd = intersect_dicts(csd, self.state_dict())  # intersect
        self.load_state_dict(csd, strict=False)  # load
        if verbose:
            LOGGER.info(f'Transferred {len(csd)}/{len(self.state_dict())} items from pretrained weights')

class LightWeightModel(BaseModel):

    default_act = nn.ReLU6()  # default activation

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__()
        self.pose_rotation = pose_rotation

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        # backbone, output size = size/4
        self.backbone = module.LightWeightBackbone(muliplier=muliplier, act=self.act)

        # mainstream, output size:
        # x1 = size/8
        # x2 = size/16
        # x3 = size/16
        self.mainstream = module.MainStreamModule(muliplier=muliplier, act=self.act)

        if pose_rotation:
            self.aux = module.AuxiliaryBackbone(int(16*muliplier), 3, muliplier=muliplier, act=self.act)
            
        output_ch = int((128 + 128 + 128)*muliplier)
        self.logit = module.FERegress(output_ch, 70*2)

        self.concat = Concat()

        initialize_weights(self)

    def forward(self, x):

        x = self.backbone(x)
        
        if self.pose_rotation:
            aux = self.aux(x)

        x = self.mainstream(x)
        
        x = self.logit(x)

        if self.pose_rotation:
            x = self.concat([x, aux])

        return x

class LightWeightModelSiLU(LightWeightModel):

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__(imgz=imgz, muliplier=muliplier, pose_rotation=pose_rotation, act=nn.SiLU())

class LightWeightPyramicModel(BaseModel):

    default_act = nn.ReLU6()  # default activation

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__()
        self.pose_rotation = pose_rotation

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        # backbone, output size = size/4
        self.backbone = module2.LightWeightBackbone(muliplier=muliplier, act=self.act)

        # mainstream, output size:
        # x1 = size/8
        # x2 = size/16
        # x3 = size/16
        self.mainstream = module2.MainStreamModule(muliplier=muliplier, act=self.act)

        if pose_rotation:
            self.aux = module2.AuxiliaryBackbone(int(32*muliplier), 3, muliplier=muliplier, act=self.act)
            
        output_ch = int((64 + 128 + 128)*muliplier)
        self.logit = module2.FERegress(output_ch, 70*2)

        self.concat = Concat()

        initialize_weights(self)

    def forward(self, x):

        x = self.backbone(x)
        
        if self.pose_rotation:
            aux = self.aux(x)

        x = self.mainstream(x)
        
        x = self.logit(x)

        if self.pose_rotation:
            x = self.concat([x, aux])

        return x

class LightWeightPyramicModelSiLU(LightWeightPyramicModel):

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__(imgz=imgz, muliplier=muliplier, pose_rotation=pose_rotation, act=nn.SiLU())

class LightWeightCSPModel(BaseModel):

    default_act = nn.ReLU6()  # default activation

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__()
        self.pose_rotation = pose_rotation

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        # backbone, output size = size/4
        self.backbone = module3.LightWeightBackbone(muliplier=muliplier, act=self.act)

        self.mainstream = module3.MainStreamModule(muliplier=muliplier, act=self.act)

        if pose_rotation:
            self.aux = module3.AuxiliaryBackbone(int(32*muliplier), 3, muliplier=muliplier, act=self.act)
            
        output_ch = int((64 + 96 + 96)*muliplier)
        self.logit = module3.FERegress(output_ch, 70*2)

        self.concat = Concat()

        initialize_weights(self)

    def forward(self, x):

        x = self.backbone(x)
        
        if self.pose_rotation:
            aux = self.aux(x)

        x = self.mainstream(x)
        
        x = self.logit(x)

        if self.pose_rotation:
            x = self.concat([x, aux])

        return x

class LightWeightCSPModelSiLU(LightWeightCSPModel):

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__(imgz=imgz, muliplier=muliplier, pose_rotation=pose_rotation, act=nn.SiLU())

class LightWeightCSPModel2(BaseModel):

    default_act = nn.ReLU6()  # default activation

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__()
        self.pose_rotation = pose_rotation

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        # backbone, output size = size/4
        self.backbone = module4.LightWeightBackbone(muliplier=muliplier, act=self.act)

        self.mainstream = module4.MainStreamModule(muliplier=muliplier, act=self.act)

        if pose_rotation:
            self.aux = module4.AuxiliaryBackbone(int(32*muliplier), 3, muliplier=muliplier, act=self.act)
            
        output_ch = int((64 + 96 + 96)*muliplier)
        self.logit = module4.FERegress(output_ch, 70*2)

        if self.pose_rotation:
            self.concat = Concat()

        initialize_weights(self)

    def forward(self, x):

        x = self.backbone(x)
        
        if self.pose_rotation:
            aux = self.aux(x)

        x = self.mainstream(x)
        
        x = self.logit(x)

        if self.pose_rotation:
            x = self.concat([x, aux])

        return x

class LightWeightCSPModel2SiLU(LightWeightCSPModel2):

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__(imgz=imgz, muliplier=muliplier, pose_rotation=pose_rotation, act=nn.SiLU())

class LightWeightDirectModel(BaseModel):

    default_act = nn.ReLU6()  # default activation

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__()
        self.pose_rotation = pose_rotation

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        # backbone, output size = size/4
        self.backbone = module5.LightWeightBackbone(muliplier=muliplier, act=self.act)

        self.mainstream = module5.MainStreamModule(muliplier=muliplier, act=self.act)

        if pose_rotation:
            self.aux = module5.AuxiliaryBackbone(int(32*muliplier), 3, muliplier=muliplier, act=self.act)
            
        self.logit = module5.FERegress(((imgz//32)**2)*int(128*muliplier), 70*2)

        if self.pose_rotation:
            self.concat = Concat()

        initialize_weights(self)

    def forward(self, x):

        x = self.backbone(x)
        
        if self.pose_rotation:
            aux = self.aux(x)

        x = self.mainstream(x)
        
        x = self.logit(x)

        if self.pose_rotation:
            x = self.concat([x, aux])

        return x

class LightWeightDirectModelSiLU(LightWeightDirectModel):

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False, act=True):
        super().__init__(imgz=imgz, muliplier=muliplier, pose_rotation=pose_rotation, act=nn.SiLU())

class LandmarkModel(BaseModel):

    def __init__(self, imgz=128, muliplier=1, pose_rotation=False):
        super().__init__()
        
        self.stack1 = nn.Sequential(
            nn.Conv2d( 3, int(32*muliplier), 3, stride=2, padding=2),
            nn.BatchNorm2d(int(32*muliplier)),
            nn.ReLU(),
            nn.Conv2d(int(32*muliplier), int(32*muliplier), 3, padding=1),
            nn.BatchNorm2d(int(32*muliplier)),
            nn.ReLU()
        )

        self.stack2 = nn.Sequential(
            nn.Conv2d(int(32*muliplier), int(64*muliplier), 3, stride=2, padding=2),
            nn.BatchNorm2d(int(64*muliplier)),
            nn.ReLU(),
            nn.Conv2d(int(64*muliplier), int(64*muliplier), 3, padding=1),
            nn.BatchNorm2d(int(64*muliplier)),
            nn.ReLU()
        )

        self.stack3 = nn.Sequential(
            nn.Conv2d(int(64*muliplier), int(128*muliplier), 3, stride=2, padding=2),
            nn.BatchNorm2d(int(128*muliplier)),
            nn.ReLU(),
            nn.Conv2d(int(128*muliplier), int(128*muliplier), 3, padding=1),
            nn.BatchNorm2d(int(128*muliplier)),
            nn.ReLU(),
            nn.Conv2d(int(128*muliplier), int(128*muliplier), 3, padding=1),
            nn.BatchNorm2d(int(128*muliplier)),
            nn.ReLU()
        )

        self.stack4 = nn.Sequential(
            nn.Conv2d(int(128*muliplier), int(256*muliplier), 3, stride=2, padding=2),
            nn.BatchNorm2d(int(256*muliplier)),
            nn.ReLU(),
            nn.Conv2d(int(256*muliplier), int(256*muliplier), 3, padding=1),
            nn.BatchNorm2d(int(256*muliplier)),
            nn.ReLU(),
            nn.Conv2d(int(256*muliplier), int(256*muliplier), 3, padding=1),
            nn.BatchNorm2d(int(256*muliplier)),
            nn.ReLU(),
            nn.AdaptiveAvgPool2d((1, 1)),
        )

        self.stack5 = nn.Sequential(
            nn.Linear(int(256*muliplier), int(192*muliplier)),
            nn.Tanh(),
            nn.Linear(int(192*muliplier), 70*2),
            nn.Tanh()
        )

        initialize_weights(self)

    def forward(self, x):
        x = self.stack1(x)
        x = self.stack2(x)
        x = self.stack3(x)
        x = self.stack4(x)
        x = x.flatten(start_dim=1)
        x = self.stack5(x)
        return x
    


import torch
import numpy as np

def nme(pred, gt, left_index=[36], right_index=[45], reduced=True):

    if len(pred.shape) == 1:
        pred_axis = pred.reshape([1, pred.shape[0]//2, 2])
    elif len(pred.shape) == 2:
        pred_axis = pred.reshape([pred.shape[0], pred.shape[1]//2, 2])
    else:
        pred_axis = pred

    if len(gt.shape) == 1:
        gt_axis = gt.reshape([1, gt.shape[0]//2, 2])
    elif len(gt.shape) == 2:
        gt_axis = gt.reshape([gt.shape[0], gt.shape[1]//2, 2])
    else:
        gt_axis = gt

    eye_span = torch.linalg.norm(gt_axis[:, left_index, :] - gt_axis[:, right_index, :], dim=(1,2))
    error = torch.linalg.norm(pred_axis - gt_axis, dim=(2))/eye_span.view(eye_span.shape[0], 1)

    if reduced:
        return error.mean(dim=1)
    else:
        return error


import math

import torch
from torch import nn

__all__ = ['autopad', 'Conv', 'DWConv', 'Concat', 'IdentifyBlock', 'C2f', 'Bottleneck', 'initialize_weights']

def initialize_weights(model):
    """Initialize model weights to random values."""
    for m in model.modules():
        t = type(m)
        if t is nn.Conv2d:
            pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
        elif t is nn.BatchNorm2d:
            m.eps = 1e-5
            m.momentum = 0.03
        elif t in [nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6, nn.SiLU]:
            m.inplace = True

def autopad(k, p=None, d=1):  # kernel, padding, dilation
    """Pad to 'same' shape outputs."""
    if d > 1:
        k = d * (k - 1) + 1 if isinstance(k, int) else [d * (x - 1) + 1 for x in k]  # actual kernel-size
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p

class Conv(nn.Module):
    """Standard convolution with args(ch_in, ch_out, kernel, stride, padding, groups, dilation, activation)."""
    default_act = nn.SiLU()  # default activation

    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, d=1, act=True):
        """Initialize Conv layer with given arguments including activation."""
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p, d), groups=g, dilation=d, bias=False)
        self.bn = nn.BatchNorm2d(c2)
        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

    def forward(self, x):
        """Apply convolution, batch normalization and activation to input tensor."""
        return self.act(self.bn(self.conv(x)))

    def forward_fuse(self, x):
        """Perform transposed convolution of 2D data."""
        return self.act(self.conv(x))
    
class DWConv(Conv):
    """Depth-wise convolution."""

    def __init__(self, c1, c2, k=1, s=1, d=1, act=True):  # ch_in, ch_out, kernel, stride, dilation, activation
        super().__init__(c1, c2, k, s, g=math.gcd(c1, c2), d=d, act=act)

class Concat(nn.Module):
    """Concatenate a list of tensors along dimension."""

    def __init__(self, dimension=1):
        """Concatenates a list of tensors along a specified dimension."""
        super().__init__()
        self.d = dimension

    def forward(self, x):
        """Forward pass for the YOLOv8 mask Proto module."""
        return torch.cat(x, self.d)


class IdentifyBlock(nn.Module):

    default_act = nn.ReLU6()  # default activation

    def __init__(self, in_ch, out_ch, expand, k=3, act=True, stride=1, dilation_rate=1):
        super().__init__()

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()
        
        expand_ch = int(expand*in_ch)

        # expansion
        self.conv1 = Conv(in_ch, expand_ch, k=1, act=self.act)

        # depthwise
        self.conv2 = DWConv(expand_ch, expand_ch, k=k, s=stride, d=dilation_rate, act=self.act)

        # squeeze
        self.conv3 = Conv(expand_ch, out_ch, k=1, act=False)

        self.need_fuse = in_ch == out_ch
        self.need_pool = stride == 2
        if self.need_fuse and self.need_pool:
            self.pool = nn.MaxPool2d(2, stride=2)

    def forward(self, x):
        
        ex = self.conv1(x)
        de = self.conv2(ex)
        sq = self.conv3(de)

        if self.need_fuse:
            if self.need_pool:
                x = self.pool(x)
            return sq + x
        else:
            return sq

class Bottleneck(nn.Module):
    """Standard bottleneck."""

    default_act = nn.ReLU6()  # default activation

    def __init__(self, c1, c2, shortcut=True, g=1, k=(3, 3), e=0.5, act=True):  # ch_in, ch_out, shortcut, groups, kernels, expand
        super().__init__()

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, k[0], 1, act=self.act)
        self.cv2 = Conv(c_, c2, k[1], 1, g=g, act=self.act)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        """'forward()' applies the YOLOv5 FPN to input data."""
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))

class C2f(nn.Module):
    """CSP Bottleneck with 2 convolutions."""

    default_act = nn.ReLU6()  # default activation

    def __init__(self, c1, c2, n=1, shortcut=False, g=1, e=0.5, act=True):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        self.c = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, 2 * self.c, 1, 1, act=self.act)
        self.cv2 = Conv((2 + n) * self.c, c2, 1, act=self.act)  # optional act=FReLU(c2)
        self.m = nn.ModuleList(Bottleneck(self.c, self.c, shortcut, g, k=((3, 3), (3, 3)), e=1.0, act=self.act) for _ in range(n))

    def forward(self, x):
        """Forward pass through C2f layer."""
        y = list(self.cv1(x).chunk(2, 1))
        y.extend(m(y[-1]) for m in self.m)
        return self.cv2(torch.cat(y, 1))

    def forward_split(self, x):
        """Forward pass using split() instead of chunk()."""
        y = list(self.cv1(x).split((self.c, self.c), 1))
        y.extend(m(y[-1]) for m in self.m)
        return self.cv2(torch.cat(y, 1))


import time
import numpy as np
from PIL import Image
from pathlib import Path
from typing import Dict, Union
from types import SimpleNamespace

from fdfat import __version__
from fdfat.utils.logger import LOGGER
from fdfat.engine.base import BaseEngine
from fdfat.utils.model_utils import preprocess
from fdfat.utils.utils import render_lmk

class PredictEngine(BaseEngine):

    def __init__(self, cfg_: Union[str, Path, Dict, SimpleNamespace]):
        super().__init__(cfg_)
        self.load_model()

        self.target_checkpoint_path = self.cfgs.checkpoint if self.cfgs.checkpoint is not None else self.save_best
        self.load_checkpoint(self.target_checkpoint_path)

        self.net.eval()

    def predict(self, input, render=False):

        if isinstance(input, str):
            input = Image.open(input)

        preprocessed = preprocess(input, self.cfgs.imgsz)
        preprocessed = torch.from_numpy(preprocessed.astype(np.float32)).to(self.cfgs.device)
        with torch.no_grad():
            if self.cfgs.test_warmup:
                for _ in range(5):
                    _ = self.net(preprocessed)
            start = time.time()
            y = self.net(preprocessed)
            y = y.detach().cpu().numpy()
            end = time.time()

        LOGGER.info(f"Predicted in {int((end-start)*1000):d}ms")

        lmk = y[0][:70*2].reshape((70,2))

        if render:
            rendered = render_lmk(input.copy(), (lmk+0.5)*input.size[0], point_size=1)
            return lmk, rendered

        return lmk


from difflib import get_close_matches
from pathlib import Path
from types import SimpleNamespace
from typing import Dict, Union

from fdfat.cfg.yaml_utils import yaml_load, yaml_save, yaml_print, DEFAULT_CFG_DICT, DEFAULT_CFG_DATA, IterableSimpleNamespace

CFG_INT_KEYS= 'seed', 'workers', 'imgsz', 'epoch', 'batch_size', 'dump_batch', 'patience', 'warmup', "lmk_num"
CFG_FRACTION_KEYS = 'lr', 'lr0_factor', 'lr_factor', 'aux_pose_weight'
CFG_FLOAT_KEYS = "aug", 'muliplier', 'w_jaw', 'w_leyeb', 'w_reyeb', 'w_nose', 'w_nosetip', 'w_leye', 'w_reye', 'w_mount', 'w_purpil'
CFG_BOOL_KEYS = "save", "override", "pin_memory", "aux_pose", "lossw_enabled", "resume", "warmup", "pre_norm", "export_simplify", "export_pose", "lmk_mean", "cache", "profiler"

def cfg2dict(cfg):
    """
    Convert a configuration object to a dictionary, whether it is a file path, a string, or a SimpleNamespace object.

    Inputs:
        cfg (str) or (Path) or (SimpleNamespace): Configuration object to be converted to a dictionary.

    Returns:
        cfg (dict): Configuration object in dictionary format.
    """
    if isinstance(cfg, (str, Path)):
        cfg = yaml_load(cfg)  # load dict
    elif isinstance(cfg, SimpleNamespace):
        cfg = vars(cfg)  # convert to dict
    return cfg

def check_cfg_mismatch(base: Dict, custom: Dict, e=None):
    """
    This function checks for any mismatched keys between a custom configuration list and a base configuration list.
    If any mismatched keys are found, the function prints out similar keys from the base list and exits the program.

    Inputs:
        - custom (Dict): a dictionary of custom configuration options
        - base (Dict): a dictionary of base configuration options
    """
    # custom = _handle_deprecation(custom)
    base, custom = (set(x.keys()) for x in (base, custom))
    mismatched = [x for x in custom if x not in base]
    if mismatched:
        string = ''
        for x in mismatched:
            matches = get_close_matches(x, base)  # key list
            matches = [f'{k}={DEFAULT_CFG_DICT[k]}' if DEFAULT_CFG_DICT.get(k) is not None else k for k in matches]
            match_str = f'Similar arguments are i.e. {matches}.' if matches else ''
            # string += f"'{colorstr('red', 'bold', x)}' is not a valid YOLO argument. {match_str}\n"
        raise SyntaxError(string) from e

def get_cfg(cfg: Union[str, Path, Dict, SimpleNamespace] = DEFAULT_CFG_DICT, overrides: Dict = None):
    """
    Load and merge configuration data from a file or dictionary.

    Args:
        cfg (str) or (Path) or (Dict) or (SimpleNamespace): Configuration data.
        overrides (str) or (Dict), optional: Overrides in the form of a file name or a dictionary. Default is None.

    Returns:
        (SimpleNamespace): Training arguments namespace.
    """
    cfg = cfg2dict(cfg)

    # Merge overrides
    if overrides:
        overrides = cfg2dict(overrides)
        check_cfg_mismatch(cfg, overrides)
        cfg = {**cfg, **overrides}  # merge cfg and overrides dicts (prefer overrides)

    # Special handling for numeric project/names
    for k in 'project', 'name':
        if k in cfg and isinstance(cfg[k], (int, float)):
            cfg[k] = str(cfg[k])

    # Type and Value checks
    for k, v in cfg.items():
        if v is not None:  # None values may be from optional args
            if k in CFG_FLOAT_KEYS and not isinstance(v, (int, float)):
                raise TypeError(f"'{k}={v}' is of invalid type {type(v).__name__}. "
                                f"Valid '{k}' types are int (i.e. '{k}=0') or float (i.e. '{k}=0.5')")
            elif k in CFG_FRACTION_KEYS:
                if not isinstance(v, (int, float)):
                    raise TypeError(f"'{k}={v}' is of invalid type {type(v).__name__}. "
                                    f"Valid '{k}' types are int (i.e. '{k}=0') or float (i.e. '{k}=0.5')")
                if not (0.0 <= v <= 1.0):
                    raise ValueError(f"'{k}={v}' is an invalid value. "
                                     f"Valid '{k}' values are between 0.0 and 1.0.")
            elif k in CFG_INT_KEYS and not isinstance(v, int):
                raise TypeError(f"'{k}={v}' is of invalid type {type(v).__name__}. "
                                f"'{k}' must be an int (i.e. '{k}=8')")
            elif k in CFG_BOOL_KEYS and not isinstance(v, bool):
                raise TypeError(f"'{k}={v}' is of invalid type {type(v).__name__}. "
                                f"'{k}' must be a bool (i.e. '{k}=True' or '{k}=False')")

    # Return instance
    return IterableSimpleNamespace(**cfg)

def get_cfg_data(cfg: Union[str, Path, Dict, SimpleNamespace] = DEFAULT_CFG_DATA, overrides: Dict = None):
    return get_cfg(cfg, overrides)


import tqdm
import torch
import numpy as np
from collections import defaultdict

from fdfat.utils.utils import render_batch
from fdfat.utils.profiler import Profile
from fdfat import TQDM_BAR_FORMAT
from fdfat.metric.metric import nme
from fdfat.data.dataloader import LMK_POINT_MEANS
from fdfat.utils.model_utils import normalize_tensor

def train_loop(cfgs, current_epoch, dataloader, model, loss_fn, optimizer, name="Train"):
    loss_dict = defaultdict(lambda: 0)
    nme_list = []

    if cfgs.lossw_enabled:
        loss_weight = torch.zeros((cfgs.batch_size, cfgs.lmk_num*2))
        all_weights = [
                cfgs.w_jaw, 
                cfgs.w_leyeb, cfgs.w_reyeb, 
                cfgs.w_nose, cfgs.w_nosetip, 
                cfgs.w_leye, cfgs.w_reye, 
                cfgs.w_mount, cfgs.w_purpil
            ]
        for idx, (b, e) in enumerate(cfgs.lmk_parts):
            loss_weight[:, b*2:e*2] = all_weights[idx]
        loss_weight = loss_weight.to(cfgs.device)

    # if cfgs.lmk_mean:
    #     lmk_means = torch.from_numpy(LMK_POINT_MEANS)

    num_batches = len(dataloader)
    pbar = tqdm.tqdm(enumerate(dataloader), total=num_batches, bar_format=TQDM_BAR_FORMAT)

    model.train()
    for batch, (x, y) in pbar:
        x_device = x.to(cfgs.device, non_blocking=True)
        y_device = y.to(cfgs.device, non_blocking=True)

        if not cfgs.pre_norm:
            x_device = normalize_tensor(x_device).type(torch.float32)

        pred = model(x_device)

        if cfgs.aux_pose:
            main_loss = loss_fn(pred[:,:cfgs.lmk_num*2], y_device[:,:cfgs.lmk_num*2])
            pose_loss = loss_fn(pred[:,-3:], y_device[:,-4:-1])

            if cfgs.lossw_enabled:
                main_loss[:, :cfgs.lmk_num*2] *= loss_weight[:main_loss.shape[0], :cfgs.lmk_num*2]

            pose_weight = y_device[:,-1:]
            pose_loss *= cfgs.aux_pose_weight * pose_weight

            total_loss = (main_loss.mean() + pose_loss.mean())/2

        else:
            main_loss = loss_fn(pred[:,:cfgs.lmk_num*2], y_device[:,:cfgs.lmk_num*2])
            total_loss = main_loss.mean()

        # Backpropagation
        total_loss.backward()
        optimizer.step()
        optimizer.zero_grad()

        loss_dict["total"] += total_loss.item()
        # if cfgs.lmk_mean:
        #     nme_val = nme(pred[:,:68*2], y_device[:,:68*2]).mean().cpu().detach().numpy()
        # else:
        nme_val = nme(pred[:,:68*2], y_device[:,:68*2]).mean().cpu().detach().numpy()
        loss_dict["nme"] += nme_val
        nme_list.append(nme_val)

        for (b, e), pname in zip(cfgs.lmk_parts, cfgs.lmk_part_names):
            loss_dict[pname] += main_loss[:, b*2:e*2].mean().item()

        losses_str = f"Total: {loss_dict['total']/(batch+1):>7f}, nmei: {loss_dict['nme']/(batch+1):>7f}"
        for n in cfgs.lmk_part_names:
            losses_str = f"{losses_str}, {n}: {loss_dict[n]/(batch+1):>7f}"

        if cfgs.aux_pose:
            loss_dict["pose"] += (pose_loss / cfgs.aux_pose_weight).mean().item()
            losses_str = f"{losses_str} pose: {loss_dict['pose']/(batch+1):>7f}"

        pbar.set_description(f"{name} epoch [{current_epoch+1:3d}/{cfgs.epoch:3d}] {losses_str}")

        if current_epoch == 0 and batch < cfgs.dump_batch:
            save_batch_png = cfgs.save_dir / f'{name}_batch{batch}.png'
            render_batch(x_device.cpu().detach().numpy(), y_device[:,:70*2].cpu().detach().numpy(), save_batch_png)

    for k in loss_dict.keys():
        loss_dict[k] /= num_batches

    nme_list = np.array(nme_list)
    loss_dict["nme_min"] = np.min(nme_list)
    loss_dict["nme_max"] = np.max(nme_list)
    loss_dict["nme_mean"] = np.mean(nme_list)
    loss_dict["nme_median"] = np.median(nme_list)
    loss_dict["nme_std"] = np.std(nme_list)

    return loss_dict


import time
from datetime import datetime
import json
from pathlib import Path
from typing import Dict, Union
from types import SimpleNamespace

import torch
from torch import nn
from torch.utils.data import DataLoader

from fdfat import __version__, MACOS, LINUX, WINDOWS
from fdfat.utils.logger import LOGGER
from fdfat.engine.base import BaseEngine
from fdfat.data.dataloader import LandmarkDataset
from fdfat.utils.utils import LMK_PART_NAMES, read_file_list, render_lmk, generate_graph, render_lmk_nme

from fdfat.engine.loop.validator import val_loop


class ValEngine(BaseEngine):

    def __init__(self, cfg_: Union[str, Path, Dict, SimpleNamespace]):
        super().__init__(cfg_)
        self.load_database()
        self.load_model()
    
    def load_database(self):

        self.target_data_path = self.cfgs.data.val
        self.target_cache_path = self.cfgs.data.val_cache
        if self.cfgs.validation == "train":
            self.target_data_path = self.cfgs.data.train
            self.target_cache_path = self.cfgs.data.train_cache
        elif self.cfgs.validation == "test":
            self.target_data_path = self.cfgs.data.test
            self.target_cache_path = self.cfgs.data.test_cache
        self.target_data_path = Path(self.target_data_path)
        self.target_cache_path = Path(self.target_cache_path) if self.target_cache_path is not None else None

        LOGGER.info(f"Load {self.cfgs.validation} data")

        self.dataset = LandmarkDataset(self.cfgs, read_file_list(self.target_data_path, base_path=self.cfgs.data.base_path), cache_path=self.target_cache_path, aug=False)
        self.dataloader = DataLoader(self.dataset, batch_size=self.cfgs.batch_size, shuffle=False,
                                        pin_memory=self.cfgs.pin_memory,
                                        num_workers=self.cfgs.workers,
                                        persistent_workers=True,
                                        multiprocessing_context="spawn" if MACOS else None)
        
        LOGGER.info("Load database DONE")

    def prepare(self):
        self.target_checkpoint_path = self.cfgs.checkpoint if self.cfgs.checkpoint is not None else self.save_best
        self.load_checkpoint(self.target_checkpoint_path)
        self.loss_fn = getattr(nn, self.cfgs.loss)(reduction='none')

    def do_validate(self, verbose=True):
        start_time = time.time()
        loss_dict = val_loop(self.cfgs, 0, self.dataloader, self.net, self.loss_fn)
        LOGGER.info(f"DONE in {int(time.time() - start_time)}s")

        if verbose:
            for k, v in loss_dict.items():
                print(k, v, "\n")

        timenow = datetime.now().isoformat()
        fname = self.target_data_path.stem
        validate_name = f"validate_{self.cfgs.validation}_{fname}_{timenow}"
        validate_record_path = self.save_dir / f"{validate_name}.json"
        validate_lmk_nme_path = self.save_dir / f"{validate_name}.jpg"

        loss_dict["date"] = timenow
        loss_dict["datapath"] = str(self.target_data_path)
        loss_dict["database"] = str(self.cfgs.data.base_path)

        with open(validate_record_path, "w") as f:
            json.dump(loss_dict, f)

        lmk_nme = render_lmk_nme(loss_dict["nme_stat"]["nme_part"]["nme_mean"], title=validate_name)
        lmk_nme.save(validate_lmk_nme_path)
        
        return loss_dict



import platform

__version__ = '0.1.17.3'

LOGGING_NAME = 'landmark'
VERBOSE = 'true'
MACOS, LINUX, WINDOWS = (platform.system() == x for x in ['Darwin', 'Linux', 'Windows'])  # environment booleans
TQDM_BAR_FORMAT = '{l_bar}{bar:10}{r_bar}'



import time
from pathlib import Path
from copy import deepcopy
from datetime import datetime
from typing import Dict, Union
from types import SimpleNamespace

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import ConstantLR, SequentialLR, LinearLR, ReduceLROnPlateau, MultiStepLR

from fdfat import __version__, MACOS, LINUX, WINDOWS
from fdfat.utils.logger import LOGGER
from fdfat.engine.base import BaseEngine
from fdfat.data.dataloader import LandmarkDataset
from fdfat.utils.utils import LMK_PART_NAMES, read_file_list, render_lmk, generate_graph

from fdfat.engine.loop.trainer import train_loop
from fdfat.engine.loop.validator import val_loop

class TrainEngine(BaseEngine):

    def __init__(self, cfg_: Union[str, Path, Dict, SimpleNamespace]):
        super().__init__(cfg_)
        self.load_database()
        self.load_model()

    def load_database(self):
        LOGGER.info("Load Train data")
        self.dataset = LandmarkDataset(self.cfgs, read_file_list(self.cfgs.data.train, base_path=self.cfgs.data.base_path), cache_path=self.cfgs.data.train_cache)
        self.train_dataloader = DataLoader(self.dataset, batch_size=self.cfgs.batch_size, shuffle=True, 
                                        pin_memory=self.cfgs.pin_memory,
                                        num_workers=self.cfgs.workers,
                                        persistent_workers=True,
                                        multiprocessing_context="spawn" if MACOS else None)

        LOGGER.info("Load Val data")
        self.dataset_test = LandmarkDataset(self.cfgs, read_file_list(self.cfgs.data.val, base_path=self.cfgs.data.base_path), cache_path=self.cfgs.data.val_cache, aug=False)
        self.test_dataloader = DataLoader(self.dataset_test, batch_size=self.cfgs.batch_size, shuffle=False,
                                        pin_memory=self.cfgs.pin_memory,
                                        num_workers=self.cfgs.workers,
                                        persistent_workers=True,
                                        multiprocessing_context="spawn" if MACOS else None)

        LOGGER.info("Load database DONE")

    def load_checkpoint(self, checkpoint):
        super().load_checkpoint(checkpoint)
        if self.optimizer is not None:
            self.optimizer.load_state_dict(self.checkpoint['optimizer']) 

    def save_model(self, epoch, save_path):
        ckpt = {
            'epoch': epoch,
            'model': deepcopy(self.net).half(),
            'best_fit': self.best_epoch_loss,
            'best_epoch': self.best_epoch_no,
            'optimizer': self.optimizer.state_dict(),
            'train_args': self.cfgs,
            'date': datetime.now().isoformat(),
            'version': __version__
        }
        torch.save(ckpt, save_path)

    def current_lr(self) -> [float]:
        lrs = []
        for i, param_group in enumerate(self.optimizer.param_groups):
            lrs.append(float(param_group['lr']))
        return lrs

    def prepare(self):
        self.loss_fn = getattr(nn, self.cfgs.loss)(reduction='none')
        if self.cfgs.optimizer == "SGD":
            self.optimizer = getattr(torch.optim, self.cfgs.optimizer)(self.net.parameters(), lr=self.cfgs.lr, momentum=0.95, nesterov=True)
        else:
            self.optimizer = getattr(torch.optim, self.cfgs.optimizer)(self.net.parameters(), lr=self.cfgs.lr)

        if self.cfgs.resume:
            self.load_checkpoint(self.save_last)
        elif self.cfgs.checkpoint is not None:
            super().load_checkpoint(self.cfgs.checkpoint, epoch_info=False) # load weight

        # scheduler_warmup = ConstantLR(self.optimizer, factor=self.cfgs.lr0_factor, total_iters=self.cfgs.warmup_epoch)
        # scheduler_main = LinearLR(self.optimizer, start_factor=1/self.cfgs.lr0_factor, end_factor=self.cfgs.lre_factor, total_iters=self.cfgs.epoch-self.cfgs.warmup_epoch)
        # self.scheduler = SequentialLR(self.optimizer, schedulers=[scheduler_warmup, scheduler_main], milestones=[self.cfgs.warmup_epoch], last_epoch=self.start_epoch)
        # self.scheduler = ReduceLROnPlateau(self.optimizer, factor=self.cfgs.lre_factor, patience=self.cfgs.patience)
        self.scheduler = MultiStepLR(self.optimizer, milestones=[100,130], gamma=self.cfgs.lre_factor)

        if not self.cfgs.resume:
            if self.cfgs.save:
                with open(self.save_log_csv, "a") as f:
                    fields = '\t'.join(LMK_PART_NAMES)
                    fields_test = '\t'.join([f"test_{a}" for a in LMK_PART_NAMES])
                    f.write(f"epoch\ttotal\tnme\t{fields}\tpose\ttest_total\ttest_nme\t{fields_test}\ttest_pose\n")

    def do_train(self):
        start_train_time = time.time()
        for current_epoch in range(self.start_epoch, self.cfgs.epoch):
            LOGGER.info(f"\n\nEPOCH {current_epoch+1}, lr: {self.current_lr()[0]:0.7f}")
            
            train_loss_dict = train_loop(self.cfgs, current_epoch, self.train_dataloader, self.net, self.loss_fn, self.optimizer)
            test_loss_dict = val_loop(self.cfgs, current_epoch, self.test_dataloader, self.net, self.loss_fn)
            self.scheduler.step()
            # self.scheduler.step(test_loss_dict['total'])

            if test_loss_dict["total"] < self.best_epoch_loss:
                self.best_epoch_loss = test_loss_dict["total"]
                self.best_epoch_no = current_epoch
            # else:
            #     if current_epoch - self.best_epoch_no > self.cfgs.patience:
            #         LOGGER.info(f"STOPPED due to no improvement after {current_epoch - self.best_epoch_no} epochs")
            #         break
            
            if self.cfgs.save:
                with open(self.save_log_csv, "a") as f:
                    f.write(f"{current_epoch+1}")
                    f.write(f"\t{train_loss_dict['total']}")
                    f.write(f"\t{train_loss_dict['nme']}")
                    for n in LMK_PART_NAMES:
                        f.write(f"\t{train_loss_dict[n]}")
                    f.write(f"\t{train_loss_dict['pose']}")

                    f.write(f"\t{test_loss_dict['total']}")
                    f.write(f"\t{test_loss_dict['nme']}")
                    for n in LMK_PART_NAMES:
                        f.write(f"\t{test_loss_dict[n]}")
                    f.write(f"\t{test_loss_dict['pose']}")
                    
                    f.write(f"\n")

                generate_graph(self.save_log_csv, self.save_log_png)

                self.save_model(current_epoch, self.save_last)
                if self.best_epoch_no == current_epoch:
                    self.save_model(current_epoch, self.save_best)
                    LOGGER.info(f"---> Saved best: epoch: {current_epoch+1}, loss: {self.best_epoch_loss:>7f}")

        LOGGER.info(f"DONE in {int(time.time() - start_train_time)}s, best epoch: {self.best_epoch_no}, val loss: {self.best_epoch_loss:>7f}")



import tqdm
import torch
import numpy as np
from collections import defaultdict

from fdfat.utils.utils import render_batch
from fdfat import TQDM_BAR_FORMAT
from fdfat.metric.metric import nme
from fdfat.utils.model_utils import normalize_tensor

def val_loop(cfgs, current_epoch, dataloader, model, loss_fn, name="Valid"):

    with torch.no_grad():
        loss_dict = defaultdict(lambda: 0)
        nme_list = []

        num_batches = len(dataloader)
        pbar = tqdm.tqdm(enumerate(dataloader), total=num_batches, bar_format=TQDM_BAR_FORMAT)

        model.eval()
        for batch, (x, y) in pbar:
            x_device = x.to(cfgs.device, non_blocking=True)
            y_device = y.to(cfgs.device, non_blocking=True)
            if not cfgs.pre_norm:
                x_device = normalize_tensor(x_device).type(torch.float32)

            pred = model(x_device)

            if cfgs.aux_pose:
                main_loss = loss_fn(pred[:,:cfgs.lmk_num*2], y_device[:,:cfgs.lmk_num*2])
                pose_loss = loss_fn(pred[:,-3:], y_device[:,-4:-1])

                pose_weight = y_device[:,-1:]
                pose_loss *= cfgs.aux_pose_weight * pose_weight

                total_loss = (main_loss.mean() + pose_loss.mean())/2
            else:
                main_loss = loss_fn(pred[:,:cfgs.lmk_num*2], y_device[:,:cfgs.lmk_num*2])
                total_loss = main_loss.mean()

            loss_dict["total"] += total_loss.item()
            nme_val = nme(pred[:,:68*2], y_device[:,:68*2], reduced=False).cpu().detach().numpy()
            loss_dict["nme"] += nme_val.mean()
            nme_list.append(nme_val)

            for (b, e), pname in zip(cfgs.lmk_parts, cfgs.lmk_part_names):
                loss_dict[pname] += main_loss[:, b*2:e*2].mean().item()

            losses_str = f"Total: {loss_dict['total']/(batch+1):>7f}, nmei: {loss_dict['nme']/(batch+1):>7f}"
            for n in cfgs.lmk_part_names:
                losses_str = f"{losses_str}, {n}: {loss_dict[n]/(batch+1):>7f}"

            if cfgs.aux_pose:
                loss_dict["pose"] += (pose_loss / cfgs.aux_pose_weight).mean().item()
                losses_str = f"{losses_str}, pose: {loss_dict['pose']/(batch+1):>7f}"

            pbar.set_description(f"{name} epoch [{current_epoch+1:3d}/{cfgs.epoch:3d}] {losses_str}")

            if current_epoch == 0 and batch < cfgs.dump_batch:
                save_batch_png = cfgs.save_dir / f'{name}_batch{batch}.png'
                render_batch(x_device.cpu().detach().numpy(), y_device[:,:70*2].cpu().detach().numpy(), save_batch_png)

        for k in loss_dict.keys():
            loss_dict[k] /= num_batches

        for k, v in loss_dict.items():
            loss_dict[k] = float(v)

        nme_list = np.concatenate(nme_list)

        loss_dict["nme_stat"] ={
            "nme_min": float(np.min(nme_list)),
            "nme_max": float(np.max(nme_list)),
            "nme_mean": float(np.mean(nme_list)),
            "nme_median": float(np.median(nme_list)),
            "nme_std": float(np.std(nme_list)),
            "nme_part": {
                "nme_min": np.min(nme_list, axis=0).tolist(),
                "nme_max": np.max(nme_list, axis=0).tolist(),
                "nme_mean": np.mean(nme_list, axis=0).tolist(),
                "nme_median": np.median(nme_list, axis=0).tolist(),
                "nme_std": np.std(nme_list, axis=0).tolist(),
            }
        }
        
        return loss_dict


import re
from pathlib import Path

import pkg_resources as pkg
from setuptools import find_packages, setup

# Settings
FILE = Path(__file__).resolve()
PARENT = FILE.parent  # root directory
README = (PARENT / 'README.md').read_text(encoding='utf-8')
REQUIREMENTS = [f'{x.name}{x.specifier}' for x in pkg.parse_requirements((PARENT / 'requirements.txt').read_text())]

def get_version():
    file = PARENT / 'fdfat/__init__.py'
    return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', file.read_text(encoding='utf-8'), re.M)[1]

setup(
    name='fdfat',  # name of pypi package
    version=get_version(),  # version of pypi package
    python_requires='>=3.7',
    license='GPL-3.0',
    description=('Fast 6DoF Face Alignment and Tracking'),
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/RyanDam/Fast-6DoF-Face-Alignment-and-Tracking',
    author='RyanDam',
    packages=find_packages(),  # required
    include_package_data=True,
    install_requires=REQUIREMENTS,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows', ],
    keywords='machine-learning, deep-learning, vision, ML, DL, AI',
    entry_points={
        'console_scripts': ['fdfat = fdfat.cli:entrypoint']})



import time
import numpy as np
from PIL import Image
from pathlib import Path
from typing import Dict, Union
from types import SimpleNamespace

from fdfat import __version__
from fdfat.utils.logger import LOGGER
from fdfat.engine.base import BaseEngine
from fdfat.utils.model_utils import preprocess, get_latest_opset
from fdfat.utils.utils import render_lmk

class ExportEngine(BaseEngine):

    def __init__(self, cfg_: Union[str, Path, Dict, SimpleNamespace]):
        super().__init__(cfg_)
        self.load_model() # prevent miss behaviour with batchnorm not in eval mode

        self.target_checkpoint_path = self.cfgs.checkpoint if self.cfgs.checkpoint is not None else self.save_best
        self.load_checkpoint(self.target_checkpoint_path)

        self.net.eval()

    def export(self):
        import torch

        if self.cfgs.export_format == "torchscript":
            target_saved_path = self.save_wdir / 'torchscript.pt'

            temp_input = torch.rand(1, 3, self.cfgs.imgsz, self.cfgs.imgsz, requires_grad=True)

            traced_net = torch.jit.trace(self.net, temp_input)
            
            traced_net.save(target_saved_path)

            LOGGER.info(f"Model saved to: {target_saved_path}")

        elif self.cfgs.export_format in ["onnx", "saved_model", "tflite"]:
            import torch.onnx
            import onnx

            onnx_saved_path = self.save_wdir / 'model.onnx'

            temp_input = torch.rand(1, 3, self.cfgs.imgsz, self.cfgs.imgsz, requires_grad=True)

            output = self.net(temp_input)

            # Export the model
            torch.onnx.export(self.net,               # model being run
                            temp_input,                         # model input (or a tuple for multiple inputs)
                            onnx_saved_path,   # where to save the model (can be a file or file-like object)
                            export_params=True,        # store the trained parameter weights inside the model file
                            do_constant_folding=True,  # whether to execute constant folding for optimization
                            opset_version=get_latest_opset(),
                            input_names = ['input'],   # the model's input names
                            output_names = ['output']) # the model's output names

            if self.cfgs.export_simplify:
                import onnxsim
                model_onnx = onnx.load(onnx_saved_path)  # load onnx model
                LOGGER.info(f'Simplifying with onnxsim {onnxsim.__version__}...')
                model_onnx, check = onnxsim.simplify(model_onnx)
                assert check, 'Simplified ONNX model could not be validated'
                onnx.save(model_onnx, onnx_saved_path)

            LOGGER.info(f"Model saved to: {onnx_saved_path}")

            if self.cfgs.export_format in ["saved_model", "tflite"]:
                import onnx2tf
                
                tf_saved_path = self.save_wdir / 'saved_model'

                onnx2tf.convert(
                    input_onnx_file_path=onnx_saved_path,
                    output_folder_path=tf_saved_path,
                    not_use_onnxsim=True,
                    non_verbose=True
                )

                LOGGER.info(f"Model saved to: {tf_saved_path}")


import torch
from torch import nn

from .conv import *
        
class LightWeightBackbone(nn.Module):

    default_act = nn.ReLU6()  # default activation

    def __init__(self, expand_rate=2, muliplier=1, act=True):
        super().__init__()

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        # early dropout
        self.conv11 = Conv(3, int(8*muliplier), k=3, s=2, act=self.act)

        self.c2f1 = C2f(int(8*muliplier), int(16*muliplier), act=self.act)
        self.c2f2 = C2f(int(16*muliplier), int(32*muliplier), act=self.act)

    def forward(self, x):

        x = self.conv11(x)
        x = self.c2f1(x)
        x = self.c2f2(x)

        return x
    
class AuxiliaryBackbone(nn.Module):

    default_act = nn.ReLU6()  # default activation

    def __init__(self, in_ch, out_ch, muliplier=1, act=True):
        super().__init__()

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        self.stack1 = nn.Sequential(
            nn.Conv2d(in_ch, int(32*muliplier), 3, 2, autopad(3)),
            nn.BatchNorm2d(int(32*muliplier)),
            self.act,
            nn.Conv2d(int(32*muliplier), int(64*muliplier), 3, 1, autopad(3)),
            nn.BatchNorm2d(int(64*muliplier)),
            self.act,
            nn.Conv2d(int(64*muliplier), int(64*muliplier), 3, 2, autopad(3)),
            nn.BatchNorm2d(int(64*muliplier)),
            self.act,
            nn.Conv2d(int(64*muliplier), int(128*muliplier), 3, 1, autopad(3)),
            nn.BatchNorm2d(int(128*muliplier)),
            self.act,
            nn.AdaptiveAvgPool2d((1, 1)),
        )

        self.stack2 = nn.Sequential(
            nn.Linear(int(128*muliplier), out_ch*3),
            nn.Tanh(),
            nn.Linear(out_ch*3, out_ch),
            nn.Tanh()
        )

    def forward(self, x):

        x = self.stack1(x)
        x = x.flatten(start_dim=1)
        x = self.stack2(x)

        return x
    
class MainStreamModule(nn.Module):
    default_act = nn.ReLU6()  # default activation

    def __init__(self, expand_rate=2, muliplier=1, act=True):
        super().__init__()

        self.act = self.default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

        self.conv11 = Conv(int(32*muliplier), int(32*muliplier), k=3, s=2, act=self.act)
        self.conv12 = C2f(int(32*muliplier), int(48*muliplier), act=self.act)

        self.conv21 = Conv(int(48*muliplier), int(48*muliplier), k=3, s=2, act=self.act)
        self.conv22 = C2f(int(48*muliplier), int(64*muliplier), act=self.act)

        self.conv31 = Conv(int(64*muliplier), int(64*muliplier), k=3, s=2, act=self.act)
        self.conv32 = C2f(int(64*muliplier), int(92*muliplier), act=self.act)

        self.conv41 = Conv(int(92*muliplier), int(92*muliplier), k=3, s=2, act=self.act)
        self.conv42 = C2f(int(92*muliplier), int(128*muliplier), act=self.act)

        # self.convp31 = Conv(int(128*muliplier), int(128*muliplier), k=3, s=2, act=self.act)
        # self.convp32 = C2f(int(96*muliplier), int(96*muliplier), act=self.act)

        # self.concat = Concat()
        # self.global_pool = nn.AdaptiveMaxPool2d((1, 1))

    def forward(self, x):
        # input: size / 4, 96/4 = 24

        x = self.conv11(x)
        # size / 8
        x = self.conv12(x)

        x = self.conv21(x)
        # size / 16
        x = self.conv22(x)

        x = self.conv31(x)
        # size / 32
        x = self.conv32(x)
        
        x = self.conv41(x)
        # size / 32
        x = self.conv42(x)

        x = x.flatten(start_dim=1)

        # x = self.concat([x1f, x2f, x3f])

        return x
    
class FERegress(nn.Module):

    def __init__(self, in_ch, out_ch):
        super().__init__()

        self.stack = nn.Sequential(
            nn.Linear(in_ch, out_ch),
            nn.Tanh()
        )

    def forward(self, x):
        x = self.stack(x)
        return x
        